from collections import defaultdict

import bblfsh

from sourced.ml.algorithms import TokenParser, NoopTokenParser
from sourced.ml.algorithms.uast_to_bag import Uast2BagBase


class FakeVocabulary:
    def __getitem__(self, item):
        return item


class UastTokens2Bag(Uast2BagBase):
    """
    Converts a UAST to a weighed bag of tokens via xpath.
    """

    XPATH = None  # Should be overridden in child class

    def __init__(self, token2index=None, token_parser=None):
        """
        :param xpath: The libuast xpath query to filter nodes with needed tokens.
        :param token2index: The mapping from tokens to bag keys. If None, no mapping is performed.
        :param token_parser: Specify token parser if you want to use a custom one. \
            :class:'NoopTokenParser' is used if it is not specified.
        """
        self._token2index = FakeVocabulary() if token2index is None else token2index
        self._token_parser = NoopTokenParser() if token_parser is None else token_parser

    @property
    def token_parser(self):
        return self._token_parser

    @property
    def token2index(self):
        return self._token2index

    def __call__(self, uast):
        """
        Converts a UAST to a weighed bag-of-words. The weights are words frequencies.
        The tokens are preprocessed by _token_parser.

        :param uast: The UAST root node.
        :return:
        """
        nodes = bblfsh.filter(uast, self.XPATH)
        bag = defaultdict(int)
        for node in nodes:
            for sub in self._token_parser.process_token(node.token):
                try:
                    bag[self._token2index[sub]] += 1
                except KeyError:
                    continue
        return bag


class UastIds2Bag(UastTokens2Bag):
    """
    Converts a UAST to a bag-of-identifiers.
    """

    XPATH = "//*[@roleIdentifier and not(@roleQualified)]"

    def __init__(self, token2index=None, token_parser=None):
        """
        :param token2index: The mapping from tokens to bag keys. If None, no mapping is performed.
        :param token_parser: Specify token parser if you want to use a custom one. \
            :class:'TokenParser' is used if it is not specified.
        """
        token_parser = TokenParser() if token_parser is None else token_parser
        super().__init__(token2index, token_parser)



import operator
from typing import Iterable

from pyspark import Row

from sourced.ml.extractors import BagsExtractor
from sourced.ml.transformers.transformer import Transformer
from sourced.ml.utils import EngineConstants


class Uast2Quant(Transformer):
    def __init__(self, extractors: Iterable[BagsExtractor], **kwargs):
        super().__init__(**kwargs)
        self.extractors = extractors
        self._levels = {}

    @property
    def levels(self):
        return self._levels

    def __call__(self, rows):
        uast_column = EngineConstants.Columns.Uast
        for i, extractor in enumerate(self.extractors):
            try:
                quantize = extractor.quantize
            except AttributeError:
                continue
            self._log.info("%s: performing quantization with %d partitions",
                           extractor.__class__.__name__, extractor.npartitions)
            items = rows \
                .flatMap(lambda row: row[uast_column]) \
                .flatMap(lambda uast: extractor.extract(uast)) \
                .reduceByKey(operator.add) \
                .map(lambda x: (x[0][0], (x[0][1], x[1]))) \
                .groupByKey().mapValues(list).collect()
            # (x[0][0], (x[0][1], x[1]))) <=> (class, (instance, frequency))
            quantize(items)
            self._levels[extractor.NAME] = extractor.levels
            self._log.info("Done, %d items", len(extractor.levels))



import itertools
import operator

from scipy import sparse
from bblfsh import Node
from pyspark import Row
from pyspark.rdd import PipelinedRDD

from sourced.ml.models import Cooccurrences, OrderedDocumentFrequencies
from sourced.ml.transformers import Transformer
from sourced.ml.utils import bblfsh_roles, EngineConstants


class CooccModelSaver(Transformer):
    def __init__(self, output, df_model: OrderedDocumentFrequencies, **kwargs):
        super().__init__(**kwargs)
        self.tokens_list = df_model.tokens()
        self.output = output
        self.df_model = df_model

    def __call__(self, sparse_matrix: PipelinedRDD):
        """
        Saves Cooccurrences asdf model to disk.

        :param sparse_matrix: rdd with 3 columns: matrix row, matrix column,  cell value. Use
            :class:`.CooccConstructor` to construct RDD from uasts.
        :return:
        """
        rows = sparse_matrix.collect()

        mat_row, mat_col, mat_weights = zip(*rows)
        tokens_num = len(self.tokens_list)

        self._log.info("Building matrix...")
        matrix = sparse.coo_matrix((mat_weights, (mat_row, mat_col)),
                                   shape=(tokens_num, tokens_num))
        Cooccurrences() \
            .construct(self.tokens_list, matrix) \
            .save(self.output, deps=(self.df_model,))


class CooccConstructor(Transformer):
    """
    Co-occurrence matrix calculation transformer.
    You can find an algorithm full description in blog.sourced.tech/posts/id2vec.
    """
    def __init__(self, token2index, token_parser, namespace="", **kwargs):
        super().__init__(**kwargs)
        self.token2index = token2index
        self.token_parser = token_parser
        self.namespace = namespace

    def _flatten_children(self, root):
        ids = []
        stack = list(root.children)
        while stack:
            node = stack.pop(0)
            if bblfsh_roles.IDENTIFIER in node.roles and \
                    bblfsh_roles.QUALIFIED not in node.roles:
                ids.append(node)
            else:
                stack.extend(node.children)
        return ids

    def _traverse_uast(self, uast: Node):
        stack = [uast]

        while stack:
            node = stack.pop(0)
            children = self._flatten_children(node)
            tokens = [token for ch in children for token in self.token_parser(ch.token)]

            if bblfsh_roles.IDENTIFIER in node.roles and \
                    bblfsh_roles.QUALIFIED not in node.roles:
                tokens.extend(self.token_parser(node.token))
            for pair in itertools.permutations(tokens, 2):
                yield pair

            stack.extend(children)

    def __call__(self, uasts):
        sparse_matrix = uasts.flatMap(self._process_row)\
            .reduceByKey(operator.add)\
            .map(lambda row: (row[0][0], row[0][1], row[1]))
        return sparse_matrix

    def _process_row(self, row: Row):
        for uast in row[EngineConstants.Columns.Uast]:
            for token1, token2 in self._traverse_uast(uast):
                try:
                    yield (self.token2index.value[self.namespace + token1],
                           self.token2index.value[self.namespace + token2]), 1
                except KeyError:
                    # Do not have token1 or token2 in the token2index map
                    pass



import functools
from py4j.java_gateway import java_import
from pyspark.sql import DataFrame


class Engine(object):
    """
    This is the entry point to any functionality exposed by the source{d}
    Engine. It contains methods to initialize the analysis on top of source
    code repositories.

    >>> from sourced.engine import Engine
    >>> repos_df = Engine(sparkSession, "/path/to/my/repositories").repositories
    >>> repos_df.show()

    :param session: spark session to use
    :type session: pyspark.sql.SparkSession
    :param repos_path: path to the folder where siva files are stored
    :type repos_path: str
    :param repos_format: format of the repositories inside the provided folder.
    It can be siva, bare or standard.
    :type repos_format: str
    :param skip_cleanup: don't delete unpacked siva files after using them
    :type skip_cleanup: bool
    """

    def __init__(self, session, repos_path, repos_format, skip_cleanup=False):
        self.session = session
        self.__jsparkSession = session._jsparkSession
        self.session.conf.set('spark.tech.sourced.engine.repositories.path', repos_path)
        self.session.conf.set('spark.tech.sourced.engine.repositories.format', repos_format)
        self.__jvm = self.session.sparkContext._gateway.jvm
        java_import(self.__jvm, 'tech.sourced.engine.Engine')
        java_import(self.__jvm, 'tech.sourced.engine.package$')

        try:
            self.__engine = self.__jvm.tech.sourced.engine.Engine.apply(self.__jsparkSession, repos_path, repos_format)
        except TypeError as e:
            if 'JavaPackage' in e.args[0]:
                raise Exception("package \"tech.sourced:engine:<version>\" cannot be found. Please, provide a jar with the package or install the package using --packages")
            else:
                raise e

        if skip_cleanup:
            self.__engine.skipCleanup(True)
        self.__implicits = getattr(getattr(self.__jvm.tech.sourced.engine, 'package$'), 'MODULE$')


    @property
    def repositories(self):
        """
        Returns a DataFrame with the data about the repositories found at
        the specified repositories path in the form of siva files.

        >>> repos_df = engine.repositories

        :rtype: RepositoriesDataFrame
        """
        return RepositoriesDataFrame(self.__engine.getRepositories(),
                                     self.session, self.__implicits)


    def blobs(self, repository_ids=[], reference_names=[], commit_hashes=[]):
        """
        Retrieves the blobs of a list of repositories, reference names and commit hashes.
        So the result will be a DataFrame of all the blobs in the given commits that are
        in the given references that belong to the given repositories.

        >>> blobs_df = engine.blobs(repo_ids, ref_names, hashes)

        Calling this function with no arguments is the same as:

        >>> engine.repositories.references.commits.first_reference_commit.tree_entries.blobs

        :param repository_ids: list of repository ids to filter by (optional)
        :type repository_ids: list of strings
        :param reference_names: list of reference names to filter by (optional)
        :type reference_names: list of strings
        :param commit_hashes: list of hashes to filter by (optional)
        :type commit_hashes: list of strings
        :rtype: BlobsDataFrame
        """
        if not isinstance(repository_ids, list):
            raise Exception("repository_ids must be a list")

        if not isinstance(reference_names, list):
            raise Exception("reference_names must be a list")

        if not isinstance(commit_hashes, list):
            raise Exception("commit_hashes must be a list")

        return BlobsDataFrame(self.__engine.getBlobs(repository_ids,
                                                  reference_names,
                                                  commit_hashes),
                              self.session,
                              self.__implicits)


    def parse_uast_node(self, data):
        """
        Parses a byte array and turns it into an UAST node.

        >>> engine.parse_uast_node(row["uast"])

        :param data: binary-encoded uast as a byte array
        :type data: byte array
        :rtype: UAST node
        """
        return self.__implicits.parseUASTNode(data)


class SourcedDataFrame(DataFrame):
    """
    Custom source{d} Engine DataFrame that contains some DataFrame overriden methods and
    utilities. This class should not be used directly, please get your SourcedDataFrames
    using the provided methods.

    :param jdf: Java DataFrame
    :type jdf: py4j.java_gateway.JavaObject
    :param session: Spark Session to use
    :type session: pyspark.sql.SparkSession
    :param implicits: Implicits object from Scala
    :type implicits: py4j.java_gateway.JavaObject
    """

    def __init__(self, jdf, session, implicits):
        DataFrame.__init__(self, jdf, session)
        self._session = session
        self._implicits = implicits


    @property
    def _engine_dataframe(self):
        return self._implicits.EngineDataFrame(self._jdf)

    def __generate_method(name):
        """
        Wraps the DataFrame's original method by name to return the derived class instance.
        """
        try:
            func = getattr(DataFrame, name)
        except AttributeError as e:
            # PySpark version is too old
            def func(self, *args, **kwargs):
                raise e
            return func
        wraps = getattr(functools, "wraps", lambda _: lambda f: f)  # py3.4+

        @wraps(func)
        def _wrapper(self, *args, **kwargs):
            dataframe = func(self, *args, **kwargs)
            if self.__class__ != SourcedDataFrame \
                    and isinstance(self, SourcedDataFrame) \
                    and isinstance(dataframe, DataFrame):
                return self.__class__(dataframe._jdf, self._session, self._implicits)
            return dataframe

        return _wrapper

    # The following code wraps all the methods of DataFrame as of 2.3
    alias = __generate_method("alias")
    checkpoint = __generate_method("checkpoint")
    coalesce = __generate_method("coalesce")
    crossJoin = __generate_method("crossJoin")
    crosstab = __generate_method("crosstab")
    describe = __generate_method("describe")
    distinct = __generate_method("distinct")
    dropDuplicates = __generate_method("dropDuplicates")
    drop_duplicates = dropDuplicates
    drop = __generate_method("drop")
    dropna = __generate_method("dropna")
    fillna = __generate_method("fillna")
    filter = __generate_method("filter")
    freqItems = __generate_method("freqItems")
    hint = __generate_method("hint")
    intersect = __generate_method("intersect")
    join = __generate_method("join")
    limit = __generate_method("limit")
    randomSplit = __generate_method("randomSplit")
    repartition = __generate_method("repartition")
    replace = __generate_method("replace")
    sampleBy = __generate_method("sampleBy")
    sample = __generate_method("sample")
    selectExpr = __generate_method("selectExpr")
    select = __generate_method("select")
    sort = __generate_method("sort")
    orderBy = sort
    sortWithinPartitions = __generate_method("sortWithinPartitions")
    subtract = __generate_method("subtract")
    summary = __generate_method("summary")
    toDF = __generate_method("toDF")
    unionByName = __generate_method("unionByName")
    union = __generate_method("union")
    where = filter
    withColumn = __generate_method("withColumn")
    withColumnRenamed = __generate_method("withColumnRenamed")
    withWatermark = __generate_method("withWatermark")

    __generate_method = staticmethod(__generate_method)  # to make IntelliSense happy


class RepositoriesDataFrame(SourcedDataFrame):
    """
    DataFrame containing repositories.
    This class should not be instantiated directly, please get your RepositoriesDataFrame using the
    provided methods.

    :param jdf: Java DataFrame
    :type jdf: py4j.java_gateway.JavaObject
    :param session: Spark Session to use
    :type session: pyspark.sql.SparkSession
    :param implicits: Implicits object from Scala
    :type implicits: py4j.java_gateway.JavaObject
    """

    def __init__(self, jdf, session, implicits):
        SourcedDataFrame.__init__(self, jdf, session, implicits)


    @property
    def references(self):
        """
        Returns the joined DataFrame of references and repositories.

        >>> refs_df = repos_df.references

        :rtype: ReferencesDataFrame
        """
        return ReferencesDataFrame(self._engine_dataframe.getReferences(),
                                   self._session, self._implicits)


    @property
    def head_ref(self):
        """
        Filters the current DataFrame references to only contain those rows whose reference is HEAD.

        >>> heads_df = repos_df.head_ref

        :rtype: ReferencesDataFrame
        """
        return self.references.ref('refs/heads/HEAD')


    @property
    def master_ref(self):
        """
        Filters the current DataFrame references to only contain those rows whose reference is master.

        >>> master_df = repos_df.master_ref

        :rtype: ReferencesDataFrame
        """
        return self.references.ref('refs/heads/master')


class ReferencesDataFrame(SourcedDataFrame):
    """
    DataFrame with references.
    This class should not be instantiated directly, please get your ReferencesDataFrame using the
    provided methods.

    :param jdf: Java DataFrame
    :type jdf: py4j.java_gateway.JavaObject
    :param session: Spark Session to use
    :type session: pyspark.sql.SparkSession
    :param implicits: Implicits object from Scala
    :type implicits: py4j.java_gateway.JavaObject
    """

    def __init__(self, jdf, session, implicits):
        SourcedDataFrame.__init__(self, jdf, session, implicits)


    @property
    def head_ref(self):
        """
        Filters the current DataFrame to only contain those rows whose reference is HEAD.

        >>> heads_df = refs_df.head_ref

        :rtype: ReferencesDataFrame
        """
        return self.ref('refs/heads/HEAD')


    @property
    def master_ref(self):
        """
        Filters the current DataFrame to only contain those rows whose reference is master.

        >>> master_df = refs_df.master_ref

        :rtype: ReferencesDataFrame
        """
        return self.ref('refs/heads/master')


    def ref(self, ref):
        """
        Filters the current DataFrame to only contain those rows whose reference is the given
        reference name.

        >>> heads_df = refs_df.ref('refs/heads/HEAD')

        :param ref: Reference to get
        :type ref: str
        :rtype: ReferencesDataFrame
        """
        return ReferencesDataFrame(self.filter(self.name == ref)._jdf,
                                   self._session, self._implicits)


    @property
    def commits(self):
        """
        Returns the current DataFrame joined with the commits DataFrame.

        >>> commits_df = refs_df.commits

        Take into account that getting all the commits will lead to a lot of repeated tree
        entries and blobs, thus making your query very slow.
        Most of the time what you probably want is to get the latest state of the files in
        a specific reference.
        You can use first_reference_commit for that purpose, which only gets the first
        commit of a reference, that is, the latest status of the reference.

        >>> commits_df = refs_df.commits.first_reference_commit

        :rtype: CommitsDataFrame
        """
        return CommitsDataFrame(self._engine_dataframe.getCommits(), self._session, self._implicits)


    @property
    def blobs(self):
        """
        Returns this DataFrame joined with the blobs DataSource.

        >>> blobs_df = refs_df.blobs

        :rtype: BlobsDataFrame
        """
        return BlobsDataFrame(self._engine_dataframe.getBlobs(), self._session, self._implicits)


class CommitsDataFrame(SourcedDataFrame):
    """
    DataFrame with commits data.
    This class should not be instantiated directly, please get your CommitsDataFrame using the
    provided methods.

    :param jdf: Java DataFrame
    :type jdf: py4j.java_gateway.JavaObject
    :param session: Spark Session to use
    :type session: pyspark.sql.SparkSession
    :param implicits: Implicits object from Scala
    :type implicits: py4j.java_gateway.JavaObject
    """

    def __init__(self, jdf, session, implicits):
        SourcedDataFrame.__init__(self, jdf, session, implicits)


    @property
    def first_reference_commit(self):
        """
        Returns a new DataFrame with only the first commit in a reference.
        Without calling this method, commits may appear multiple times in your DataFrame,
        because most of your commits will be shared amongst references. Calling this, your
        DataFrame will only contain the HEAD commit of each reference.

        For the next example, consider we have a master branch with 100 commits and a "foo" branch
        whose parent is the HEAD of master and has two more commits.

        >>> > commits_df.count()
        >>> 102
        >>> > commits_df.first_reference_commit.count()
        >>> 2

        :rtype: CommitsDataFrame
        """
        return CommitsDataFrame(self._engine_dataframe.getFirstReferenceCommit(), self._session,
                                self._implicits)


    @property
    def tree_entries(self):
        """
        Returns this DataFrame joined with the tree entries DataSource.

        >>> entries_df = commits_df.tree_entries

        :rtype: TreeEntriesDataFrame
        """
        return TreeEntriesDataFrame(self._engine_dataframe.getTreeEntries(), self._session, self._implicits)


    @property
    def blobs(self):
        """
        Returns a new DataFrame with the blob associated to each tree entry of the commit.

        >>> > blobs_df = commits_df.blobs

        :rtype: BlobsDataFrame
        """
        return BlobsDataFrame(self._engine_dataframe.getBlobs(), self._session,
                                self._implicits)


class TreeEntriesDataFrame(SourcedDataFrame):
    """
    DataFrame with tree entries data data.
    This class should not be instantiated directly, please get your TreeEntriesDataFrame using the
    provided methods.

    :param jdf: Java DataFrame
    :type jdf: py4j.java_gateway.JavaObject
    :param session: Spark Session to use
    :type session: pyspark.sql.SparkSession
    :param implicits: Implicits object from Scala
    :type implicits: py4j.java_gateway.JavaObject
    """

    def __init__(self, jdf, session, implicits):
        SourcedDataFrame.__init__(self, jdf, session, implicits)


    @property
    def blobs(self):
        """
        Returns a new DataFrame with the blob associated to each tree entry.

        >>> > blobs_df = trees_df.blobs

        :rtype: BlobsDataFrame
        """
        return BlobsDataFrame(self._engine_dataframe.getBlobs(), self._session,
                              self._implicits)


class BlobsDataFrame(SourcedDataFrame):
    """
    DataFrame containing blobs data.
    This class should not be instantiated directly, please get your BlobsDataFrame using the
    provided methods.

    :param jdf: Java DataFrame
    :type jdf: py4j.java_gateway.JavaObject
    :param session: Spark Session to use
    :type session: pyspark.sql.SparkSession
    :param implicits: Implicits object from Scala
    :type implicits: py4j.java_gateway.JavaObject
    """

    def __init__(self, jdf, session, implicits):
        SourcedDataFrame.__init__(self, jdf, session, implicits)


    def classify_languages(self):
        """
        Returns a new DataFrame with the language data of any blob added to
        its row.

        >>> blobs_lang_df = blobs_df.classify_languages

        :rtype: BlobsWithLanguageDataFrame
        """
        return BlobsWithLanguageDataFrame(self._engine_dataframe.classifyLanguages(),
                                          self._session, self._implicits)


    def extract_uasts(self):
        """
        Returns a new DataFrame with the parsed UAST data of any blob added to
        its row.

        >>> blobs_df.extract_uasts

        :rtype: UASTsDataFrame
        """
        return UASTsDataFrame(self._engine_dataframe.extractUASTs(),
                              self._session, self._implicits)


class BlobsWithLanguageDataFrame(SourcedDataFrame):
    """
    DataFrame containing blobs and language data.
    This class should not be instantiated directly, please get your BlobsWithLanguageDataFrame
    using the provided methods.

    :param jdf: Java DataFrame
    :type jdf: py4j.java_gateway.JavaObject
    :param session: Spark Session to use
    :type session: pyspark.sql.SparkSession
    :param implicits: Implicits object from Scala
    :type implicits: py4j.java_gateway.JavaObject
    """

    def __init__(self, jdf, session, implicits):
        SourcedDataFrame.__init__(self, jdf, session, implicits)


    def extract_uasts(self):
        """
        Returns a new DataFrame with the parsed UAST data of any blob added to
        its row.

        >>> blobs_lang_df.extract_uasts

        :rtype: UASTsDataFrame
        """
        return UASTsDataFrame(self._engine_dataframe.extractUASTs(),
                              self._session, self._implicits)


class UASTsDataFrame(SourcedDataFrame):
    """
    DataFrame containing UAST data.
    This class should not be instantiated directly, please get your UASTsDataFrame using the
    provided methods.

    :param jdf: Java DataFrame
    :type jdf: py4j.java_gateway.JavaObject
    :param session: Spark Session to use
    :type session: pyspark.sql.SparkSession
    :param implicits: Implicits object from Scala
    :type implicits: py4j.java_gateway.JavaObject
    """

    def __init__(self, jdf, session, implicits):
        SourcedDataFrame.__init__(self, jdf, session, implicits)


    def query_uast(self, query, query_col='uast', output_col='result'):
        """
        Queries the UAST of a file with the given query to get specific nodes.

        >>> rows = uasts_df.query_uast('//*[@roleIdentifier]').collect()
        >>> rows = uasts_df.query_uast('//*[@roleIdentifier]', 'foo', 'bar')

        :param query: xpath query
        :type query: str
        :param query_col: column containing the list of nodes to query
        :type query_col: str
        :param output_col: column to place the result of the query
        :type output_col: str
        :rtype: UASTsDataFrame
        """
        return UASTsDataFrame(self._engine_dataframe.queryUAST(query,
                                       query_col,
                                       output_col),
                              self._session, self._implicits)


    def extract_tokens(self, input_col='result', output_col='tokens'):
        """
        Extracts the tokens from UAST nodes.

        >>> rows = uasts_df.query_uast('//*[@roleIdentifier]').extract_tokens().collect()
        >>> rows = uasts_df.query_uast('//*[@roleIdentifier]', output_col='foo').extract_tokens('foo', 'bar')

        :param input_col: column containing the list of nodes to extract tokens from
        :type input_col: str
        :param output_col: column to place the resultant tokens
        :type output_col: str
        :rtype: UASTsDataFrame
        """
        return UASTsDataFrame(self._engine_dataframe.extractTokens(input_col, output_col),
                              self._session, self._implicits)



import re

import Stemmer


class TokenParser:
    """
    Common utilities for splitting and stemming tokens.
    """
    NAME_BREAKUP_RE = re.compile(r"[^a-zA-Z]+")  #: Regexp to split source code identifiers.
    STEM_THRESHOLD = 6  #: We do not stem splitted parts shorter than or equal to this size.
    MAX_TOKEN_LENGTH = 256  #: We cut identifiers longer than thi value.

    def __init__(self, stem_threshold=STEM_THRESHOLD, max_token_length=MAX_TOKEN_LENGTH):
        self._stemmer = Stemmer.Stemmer("english")
        self._stemmer.maxCacheSize = 0
        self._stem_threshold = stem_threshold
        self._max_token_length = max_token_length

    def __call__(self, token):
        return self.process_token(token)

    def process_token(self, token):
        for word in self.split(token):
            yield self.stem(word)

    def stem(self, word):
        if len(word) <= self._stem_threshold:
            return word
        return self._stemmer.stemWord(word)

    def split(self, token):
        token = token.strip()[:self._max_token_length]
        prev_p = [""]

        def ret(name):
            r = name.lower()
            if len(name) >= 3:
                yield r
                if prev_p[0]:
                    yield prev_p[0] + r
                    prev_p[0] = ""
            else:
                prev_p[0] = r

        for part in self.NAME_BREAKUP_RE.split(token):
            if not part:
                continue
            prev = part[0]
            pos = 0
            for i in range(1, len(part)):
                this = part[i]
                if prev.islower() and this.isupper():
                    yield from ret(part[pos:i])
                    pos = i
                elif prev.isupper() and this.islower():
                    if 0 < i - 1 - pos <= 3:
                        yield from ret(part[pos:i - 1])
                        pos = i - 1
                    elif i - 1 > pos:
                        yield from ret(part[pos:i])
                        pos = i
                prev = this
            last = part[pos:]
            if last:
                yield from ret(last)

    def __getstate__(self):
        state = self.__dict__.copy()
        del state["_stemmer"]
        return state

    def __setstate__(self, state):
        self.__dict__ = state
        self._stemmer = Stemmer.Stemmer("english")


class NoopTokenParser:
    """
    One can use this class if he or she does not want to do any parsing.
    """

    def process_token(self, token):
        yield token

    def __call__(self, token):
        return self.process_token(token)



import logging
from http.server import HTTPServer, SimpleHTTPRequestHandler, test
import os
import shutil
import threading
import time


class CORSWebServer(object):
    def __init__(self):
        self.thread = None
        self.server = None

    def serve(self):
        outer = self

        class ClojureServer(HTTPServer):
            def __init__(self, *args, **kwargs):
                HTTPServer.__init__(self, *args, **kwargs)
                outer.server = self

        class CORSRequestHandler(SimpleHTTPRequestHandler):
            def end_headers(self):
                self.send_header("Access-Control-Allow-Origin", "*")
                SimpleHTTPRequestHandler.end_headers(self)

        test(CORSRequestHandler, ClojureServer)

    def start(self):
        self.thread = threading.Thread(target=self.serve)
        self.thread.start()

    def stop(self):
        if self.running:
            self.server.shutdown()
            self.server.server_close()
            self.thread.join()
            self.server = None
            self.thread = None

    @property
    def running(self):
        return self.server is not None


web_server = CORSWebServer()


def present_embeddings(destdir, run_server, labels, index, embeddings):
    log = logging.getLogger("projector")
    log.info("Writing Tensorflow Projector files...")
    if not os.path.isdir(destdir):
        os.makedirs(destdir)
    os.chdir(destdir)
    metaf = "id2vec_meta.tsv"
    with open(metaf, "w") as fout:
        if len(labels) > 1:
            fout.write("\t".join(labels) + "\n")
        for item in index:
            if len(labels) > 1:
                fout.write("\t".join(item) + "\n")
            else:
                fout.write(item + "\n")
        log.info("Wrote %s", metaf)
    dataf = "id2vec_data.tsv"
    with open(dataf, "w") as fout:
        for vec in embeddings:
            fout.write("\t".join(str(v) for v in vec))
            fout.write("\n")
        log.info("Wrote %s", dataf)
    jsonf = "id2vec.json"
    with open(jsonf, "w") as fout:
        fout.write("""{
  "embeddings": [
    {
      "tensorName": "id2vec",
      "tensorShape": [%s, %s],
      "tensorPath": "http://0.0.0.0:8000/%s",
      "metadataPath": "http://0.0.0.0:8000/%s"
    }
  ]
}
""" % (len(embeddings), len(embeddings[0]), dataf, metaf))
    log.info("Wrote %s", jsonf)
    if run_server and not web_server.running:
        web_server.start()
    url = "http://projector.tensorflow.org/?config=http://0.0.0.0:8000/" + jsonf
    log.info(url)
    if run_server:
        if shutil.which("xdg-open") is not None:
            os.system("xdg-open " + url)
        else:
            browser = os.getenv("BROWSER", "")
            if browser:
                os.system(browser + " " + url)
            else:
                print("\t" + url)


def wait():
    log = logging.getLogger("projector")
    secs = int(os.getenv("PROJECTOR_SERVER_TIME", "60"))
    log.info("Sleeping for %d seconds, safe to Ctrl-C" % secs)
    try:
        time.sleep(secs)
    except KeyboardInterrupt:
        pass
    web_server.stop()



import modelforge.dump


def dump_model(args):
    """
    Proxy for :func:`modelforge.dump.dump_model`.
    """
    if args.gcs_bucket:
        args.args = "bucket=" + args.gcs_bucket
        del args.gcs_bucket
    else:
        args.args = ""
    args.backend = None
    return modelforge.dump.dump_model(args)



import functools
import logging
import os
import sys

os.environ["PYSPARK_PYTHON"] = sys.executable

import pyspark  # nopep8
from pyspark.sql import SparkSession  # nopep8
from sourced.engine import Engine  # nopep8


class SparkDefault:
    """
    Default arguments for create_spark function and __main__
    """
    MASTER_ADDRESS = "local[*]"
    LOCAL_DIR = "/tmp/spark"
    LOG_LEVEL = "WARN"
    CONFIG = []
    PACKAGES = []


def add_spark_args(my_parser, default_packages=None):
    my_parser.add_argument(
        "-s", "--spark", default=SparkDefault.MASTER_ADDRESS,
        help="Spark's master address.")
    my_parser.add_argument(
        "--config", nargs="+", default=SparkDefault.CONFIG,
        help="Spark configuration (key=value).")
    my_parser.add_argument(
        "-m", "--memory",
        help="Handy memory config for spark. -m 4G,10G,2G is equivalent to "
             "--config spark.executor.memory=4G "
             "--config spark.driver.memory=10G "
             "--config spark.driver.maxResultSize=2G."
             "Numbers are floats separated by commas.")
    if default_packages is None:
        default_packages = SparkDefault.PACKAGES
    my_parser.add_argument(
        "--package", nargs="+", default=default_packages, dest="packages",
        help="Additional Spark packages.")
    my_parser.add_argument(
        "--spark-local-dir", default=SparkDefault.LOCAL_DIR,
        help="Spark local directory.")
    my_parser.add_argument("--spark-log-level", default=SparkDefault.LOG_LEVEL, choices=(
        "ALL", "DEBUG", "ERROR", "FATAL", "INFO", "OFF", "TRACE", "WARN"),
                           help="Spark log level")
    persistences = [att for att in pyspark.StorageLevel.__dict__.keys() if "__" not in att]
    my_parser.add_argument(
        "--persist", default=None, choices=persistences,
        help="Spark persistence type (StorageLevel.*).")


class EngineConstants:
    """
    Constants for Engine usage.
    """
    class Columns:
        """
        Column names constants.
        """
        RepositoryId = "repository_id"
        Path = "path"
        BlobId = "blob_id"
        Uast = "uast"


class EngineDefault:
    """
    Default arguments for create_engine function and __main__
    """
    BBLFSH = "localhost"
    VERSION = "0.5.1"


def add_engine_args(my_parser, default_packages=None):
    add_spark_args(my_parser, default_packages=default_packages)
    my_parser.add_argument(
        "--bblfsh", default=EngineDefault.BBLFSH,
        help="Babelfish server's address.")
    my_parser.add_argument(
        "--engine", default=EngineDefault.VERSION,
        help="source{d} engine version.")
    my_parser.add_argument("--repository-format", default="siva",
                           help="Repository storage input format.")
    my_parser.add_argument("--explain", action="store_true",
                           help="Print the PySpark execution plans.")


def create_spark(session_name,
                 spark=SparkDefault.MASTER_ADDRESS,
                 spark_local_dir=SparkDefault.LOCAL_DIR,
                 config=SparkDefault.CONFIG,
                 packages=SparkDefault.PACKAGES,
                 spark_log_level=SparkDefault.LOG_LEVEL,
                 **_):  # **kwargs are discarded for convenience
    log = logging.getLogger("spark")
    log.info("Starting %s on %s", session_name, spark)
    builder = SparkSession.builder.master(spark).appName(session_name)
    builder = builder.config(
        "spark.jars.packages", ",".join(packages))
    builder = builder.config("spark.local.dir", spark_local_dir)
    for cfg in config:
        builder = builder.config(*cfg.split("=", 1))
    session = builder.getOrCreate()
    session.sparkContext.setLogLevel(spark_log_level)
    # Hide py4j verbose logging (It appears in travis mostly)
    logging.getLogger("py4j").setLevel(logging.WARNING)
    return session


def create_engine(session_name, repositories,
                  bblfsh=EngineDefault.BBLFSH,
                  engine=EngineDefault.VERSION,
                  config=None, packages=None, memory="",
                  repository_format="siva", **spark_kwargs):
    if config is None:
        config = []
    if packages is None:
        packages = []
    config.append("spark.tech.sourced.bblfsh.grpc.host=" + bblfsh)
    # TODO(vmarkovtsev): figure out why is this option needed
    config.append("spark.tech.sourced.engine.cleanup.skip=true")
    packages.append("tech.sourced:engine:" + engine)
    memory_conf = []
    if memory:
        memory = memory.split(",")
        err = "Expected 3 memory parameters but got {}. " \
              "Please check --help to see how -m/--memory should be used."
        assert len(memory) == 3, err.format(len(memory))
        memory_conf.append("spark.executor.memory=" + memory[0])
        memory_conf.append("spark.driver.memory=" + memory[1])
        memory_conf.append("spark.driver.maxResultSize=" + memory[2])
    session = create_spark(session_name, config=config + memory_conf, packages=packages,
                           **spark_kwargs)
    log = logging.getLogger("engine")
    log.info("Initializing on %s", repositories)
    engine = Engine(session, repositories, repository_format)
    return engine


def pause(func):
    @functools.wraps(func)
    def wrapped_pause(cmdline_args, *args, **kwargs):
        try:
            return func(cmdline_args, *args, **kwargs)
        finally:
            if cmdline_args.pause:
                input("Press Enter to exit...")

    return wrapped_pause


def pipeline_graph(args, log, root):
    if args.graph:
        log.info("Dumping the graph to %s", args.graph)
        with open(args.graph, "w") as f:
            root.graph(stream=f)



import sys

from modelforge.logs import setup_logging


utmain = sys.modules['__main__']
if utmain.__package__ == "unittest" and utmain.__spec__ is None:
    from collections import namedtuple
    ModuleSpec = namedtuple("ModuleSpec", ["name"])
    utmain.__spec__ = ModuleSpec("unittest.__main__")
    del ModuleSpec
del utmain


def setup():
    setup_logging("INFO")



from sourced.ml.utils.bigartm import install_bigartm
from sourced.ml.utils.engine import create_engine, create_spark, add_engine_args, add_spark_args, \
    EngineConstants
from sourced.ml.utils.pickleable_logger import PickleableLogger



from typing import Iterable, Dict, List

import numpy

from modelforge import register_model, merge_strings, split_strings
from sourced.ml.models import DocumentFrequencies


@register_model
class OrderedDocumentFrequencies(DocumentFrequencies):
    """
    Compatible with the original DocumentFrequencies. This model maintains the determinitic
    sequence of the tokens.
    """
    # NAME is the same

    def construct(self, docs: int, tokfreqs: Iterable[Dict[str, int]]):
        super().construct(docs, tokfreqs)
        self._log.info("Ordering the keys...")
        keys = sorted(self._df)
        self._order = {k: i for i, k in enumerate(keys)}
        return self

    @property
    def order(self) -> Dict[str, int]:
        return self._order

    def tokens(self) -> List[str]:
        arr = [None for _ in range(len(self))]
        for k, v in self.order.items():
            arr[v] = k
        return arr

    def _load_tree(self, tree):
        tokens = split_strings(tree["tokens"])
        super()._load_tree(tree, tokens)
        self._log.info("Mapping the keys order...")
        self._order = {k: i for i, k in enumerate(tokens)}

    def _generate_tree(self):
        tokens = [None] * len(self)
        freqs = numpy.zeros(len(self), dtype=numpy.float32)
        for k, i in self._order.items():
            tokens[i] = k
            freqs[i] = self._df[k]
        return {"docs": self.docs, "tokens": merge_strings(tokens), "freqs": freqs}

    def prune(self, threshold: int) -> "OrderedDocumentFrequencies":
        pruned = super().prune(threshold)
        if pruned is not self:
            self._log.info("Recovering the order...")
            pruned._order = {k: i for i, k in enumerate(sorted(pruned._df))}
        return pruned

    def greatest(self, max_size: int) -> "OrderedDocumentFrequencies":
        pruned = super().greatest(max_size)
        if pruned is not self:
            self._log.info("Recovering the order...")
            pruned._order = {k: i for i, k in enumerate(sorted(pruned._df))}
        return pruned



from sourced.ml.algorithms.tf_idf import log_tf_log_idf
from sourced.ml.algorithms.token_parser import TokenParser, NoopTokenParser
from sourced.ml.algorithms.uast_struct_to_bag import UastRandomWalk2Bag, UastSeq2Bag
from sourced.ml.algorithms.uast_inttypes_to_nodes import Uast2QuantizedChildren
from sourced.ml.algorithms.uast_inttypes_to_graphlets import Uast2GraphletBag



import logging
from uuid import uuid4

from sourced.ml.extractors import create_extractors_from_args
from sourced.ml.models import OrderedDocumentFrequencies, QuantizationLevels
from sourced.ml.transformers import Ignition, UastExtractor, UastDeserializer, \
    BagFeatures2DocFreq, HeadFiles, Uast2BagFeatures, Counter, Cacher, Uast2Quant, UastRow2Document
from sourced.ml.utils import create_engine
from sourced.ml.utils.engine import pipeline_graph, pause


@pause
def repos2df_entry(args):
    log = logging.getLogger("repos2df")
    engine = create_engine("repos2df-%s" % uuid4(), **args.__dict__)
    extractors = create_extractors_from_args(args)

    ignition = Ignition(engine, explain=args.explain)
    uast_extractor = ignition \
        .link(HeadFiles()) \
        .link(UastExtractor(languages=args.languages)) \
        .link(UastRow2Document()) \
        .link(Cacher.maybe(args.persist))
    log.info("Extracting UASTs...")
    ndocs = uast_extractor.link(Counter()).execute()
    log.info("Number of documents: %d", ndocs)
    uast_extractor = uast_extractor.link(UastDeserializer())
    quant = Uast2Quant(extractors)
    uast_extractor.link(quant).execute()
    if quant.levels:
        log.info("Writing quantization levels to %s", args.quant)
        QuantizationLevels().construct(quant.levels).save(args.quant)
    df = uast_extractor \
        .link(Uast2BagFeatures(extractors)) \
        .link(BagFeatures2DocFreq()) \
        .execute()
    log.info("Writing %s", args.docfreq)
    OrderedDocumentFrequencies().construct(ndocs, df).save(args.docfreq)
    pipeline_graph(args, log, ignition)



import argparse
import logging
from uuid import uuid4

from sourced.ml.extractors import create_extractors_from_args
from sourced.ml.models import OrderedDocumentFrequencies, QuantizationLevels
from sourced.ml.transformers import Ignition, UastExtractor, UastDeserializer, Uast2Quant, \
    BagFeatures2DocFreq, BagFeatures2TermFreq, Uast2BagFeatures, HeadFiles, TFIDF, Cacher, \
    Indexer, UastRow2Document, BOWWriter, Moder
from sourced.ml.utils import create_engine
from sourced.ml.utils.engine import pipeline_graph, pause


def add_bow_args(my_parser: argparse.ArgumentParser):
    my_parser.add_argument(
        "--bow", required=True, help="[OUT] The path to the Bag-Of-Words model.")
    my_parser.add_argument(
        "--batch", default=BOWWriter.DEFAULT_CHUNK_SIZE, type=int,
        help="The maximum size of a single BOW file in bytes.")


@pause
def repos2bow_entry_template(args, select=HeadFiles, cache_hook=None):
    log = logging.getLogger("repos2bow")
    engine = create_engine("repos2bow-%s" % uuid4(), **args.__dict__)
    extractors = create_extractors_from_args(args)

    ignition = Ignition(engine, explain=args.explain)
    uast_extractor = ignition \
        .link(select()) \
        .link(UastExtractor(languages=args.languages)) \
        .link(Moder(args.mode)) \
        .link(Cacher.maybe(args.persist))
    if cache_hook is not None:
        uast_extractor.link(cache_hook()).execute()
    # We link UastRow2Document after Cacher here because cache_hook() may want to have all possible
    # Row items.
    uast_extractor = uast_extractor.link(UastRow2Document())
    log.info("Extracting UASTs and indexing documents...")
    document_indexer = Indexer(Uast2BagFeatures.Columns.document)
    uast_extractor.link(document_indexer).execute()
    ndocs = len(document_indexer)
    log.info("Number of documents: %d", ndocs)
    uast_extractor = uast_extractor.link(UastDeserializer())
    quant = Uast2Quant(extractors)
    uast_extractor.link(quant).execute()
    if quant.levels:
        log.info("Writing quantization levels to %s", args.quant)
        QuantizationLevels().construct(quant.levels).save(args.quant)
    uast_extractor = uast_extractor \
        .link(Uast2BagFeatures(extractors)) \
        .link(Cacher.maybe(args.persist))
    log.info("Calculating the document frequencies...")
    df = uast_extractor.link(BagFeatures2DocFreq()).execute()
    log.info("Writing docfreq to %s", args.docfreq)
    df_model = OrderedDocumentFrequencies() \
        .construct(ndocs, df) \
        .prune(args.min_docfreq) \
        .greatest(args.vocabulary_size) \
        .save(args.docfreq)
    uast_extractor \
        .link(BagFeatures2TermFreq()) \
        .link(TFIDF(df_model)) \
        .link(document_indexer) \
        .link(Indexer(Uast2BagFeatures.Columns.token, df_model.order)) \
        .link(BOWWriter(document_indexer, df_model, args.bow, args.batch)) \
        .execute()
    pipeline_graph(args, log, ignition)


def repos2bow_entry(args):
    return repos2bow_entry_template(args)



from itertools import chain

from pyspark import Row, RDD
from pyspark.sql import DataFrame

from sourced.ml.transformers import Transformer
from sourced.ml.utils import EngineConstants


class Moder(Transformer):
    """
    Select the items to extract from UASTs.
    """
    class Options:
        repo = "repo"
        file = "file"
        function = "func"

        __all__ = (file, function, repo)

    # Copied from https://github.com/src-d/hercules/blob/master/shotness.go#L40
    # If you change here, please PR it to Hercules as well
    FUNC_XPATH = "//*[@roleFunction and @roleDeclaration]"
    FUNC_NAME_XPATH = "/*[@roleFunction and @roleIdentifier and @roleName] " \
                      "| /*/*[@roleFunction and @roleIdentifier and @roleName]"

    def __init__(self, mode: str, **kwargs):
        super().__init__(**kwargs)
        self.mode = mode

    def __setstate__(self, state):
        super().__setstate__(state)
        from bblfsh import Node, filter as filter_uast
        self.parse_uast = Node.FromString
        self.serialize_uast = Node.SerializeToString
        self.filter_uast = filter_uast

    @property
    def mode(self):
        return self._mode

    @mode.setter
    def mode(self, value: str):
        if not isinstance(value, str):
            raise TypeError("mode must be a string")
        if value not in self.Options.__all__:
            raise ValueError("Unsupported mode: " + value)
        self._mode = value

    def call_repo(self, rows: RDD):
        ridcol = EngineConstants.Columns.RepositoryId
        uastcol = EngineConstants.Columns.Uast
        return rows \
            .groupBy(lambda r: r[ridcol]) \
            .map(lambda x: Row(**{ridcol: x[0], EngineConstants.Columns.Path: "",
                                  EngineConstants.Columns.BlobId: "",
                                  uastcol: list(chain.from_iterable(i[uastcol] for i in x[1]))}))

    def call_file(self, rows: RDD):
        return rows

    def call_func(self, rows: RDD):
        return rows.flatMap(self.extract_functions_from_row)

    def __call__(self, rows: DataFrame) -> RDD:
        return getattr(self, "call_" + self.mode)(rows.rdd)

    def extract_functions_from_row(self, row: Row):
        uastbytes = row[EngineConstants.Columns.Uast]
        if not uastbytes:
            return
        uast = self.parse_uast(uastbytes[0])
        template = row.asDict()
        for func, name in self.extract_functions_from_uast(uast):
            data = template.copy()
            data[EngineConstants.Columns.Uast] = [self.serialize_uast(func)]
            data[EngineConstants.Columns.BlobId] += "_%s:%d" % (name, func.start_position.line)
            yield Row(**data)

    def extract_functions_from_uast(self, uast: "bblfsh.Node"):
        allfuncs = list(self.filter_uast(uast, self.FUNC_XPATH))
        internal = set()
        for func in allfuncs:
            if id(func) in internal:
                continue
            for sub in self.filter_uast(func, self.FUNC_XPATH):
                if sub != func:
                    internal.add(id(sub))
        for f in allfuncs:
            if id(f) not in internal:
                name = "+".join(n.token for n in self.filter_uast(f, self.FUNC_NAME_XPATH))
                if name:
                    yield f, name



import argparse
import inspect
from typing import List

from sourced.ml.extractors.bags_extractor import BagsExtractor

__extractors__ = {}


def register_extractor(cls):
    if not issubclass(cls, BagsExtractor):
        raise TypeError("%s is not an instance of %s" % (cls.__name__, BagsExtractor.__name__))
    __extractors__[cls.NAME] = cls
    return cls


def get_names_from_kwargs(f):
    for k, v in inspect.signature(f).parameters.items():
        if v.default != inspect.Parameter.empty and isinstance(
                v.default, (str, int, float, tuple)):
            yield k.replace("_", "-"), v.default


def filter_kwargs(kwargs, func):
    func_param = inspect.signature(func).parameters.keys()
    return dict([(k, v) for k, v in kwargs.items() if k in func_param])


def create_extractors_from_args(args: argparse.Namespace) -> List[BagsExtractor]:
    return [__extractors__[s](args.min_docfreq, log_level=args.log_level,
                              **__extractors__[s].get_kwargs_fromcmdline(args))
            for s in args.feature]



import operator

from pyspark import RDD

from sourced.ml.transformers import Transformer


class BagFeatures2DocFreq(Transformer):
    def __call__(self, rows: RDD):
        head = rows \
            .map(lambda x: x[0]) \
            .distinct() \
            .map(lambda x: (x[0], 1)) \
            .reduceByKey(operator.add)
        if self.explained:
            self._log.info("toDebugString():\n%s", head.toDebugString().decode())
        return head.collectAsMap()



from sourced.ml.cmd_entries.bigartm2asdf import bigartm2asdf_entry
from sourced.ml.cmd_entries.bow_converters import bow2vw_entry
from sourced.ml.cmd_entries.dump_model import dump_model
from sourced.ml.cmd_entries.postprocess_id2vec import postprocess_id2vec
from sourced.ml.cmd_entries.preprocess_id2vec import preprocess_id2vec
from sourced.ml.cmd_entries.projector_entry import projector_entry
from sourced.ml.cmd_entries.repos2bow import repos2bow_entry
from sourced.ml.cmd_entries.repos2coocc import repos2coocc_entry
from sourced.ml.cmd_entries.repos2df import repos2df_entry
from sourced.ml.cmd_entries.run_swivel import run_swivel



import glob
import logging
import multiprocessing
import os
import shutil
import subprocess
import tempfile


def execute(cmd, cwd, log):
    log.info(">>> %s", cmd)
    parsed = [v for v in cmd.split(" ") if v]
    subprocess.check_call(parsed, cwd=cwd)


def install_bigartm(args=None, target="./bigartm", tempdir=None, warn_exists=True):
    """
    Deploys bigartm/bigartm at the specified path.

    :param args: :class:`argparse.Namespace` with "output" and "tmpdir". \
                 "output" sets the target directory, "tmpdir" sets \
                 the temporary directory which is used to clone bigartm/bigartm \
                 and build it.
    :param target: The path to the built executable. If args is not None, it \
                   becomes overridden.
    :param tempdir: The temporary directory where to clone and build \
                    bigartm/bigartm. If args is not None, it becomes overridden.
    :return: None if successful; otherwise, the error code (can be 0!).
    """
    log = logging.getLogger("bigartm")
    if args is not None:
        tempdir = args.tmpdir
        target = os.path.join(args.output, "bigartm")
    if shutil.which(os.path.basename(target)) or shutil.which(target, path=os.getcwd()):
        if warn_exists:
            log.warning("bigartm is in the PATH, no-op.")
        return 0
    if not shutil.which("cmake"):
        log.error("You need to install cmake.")
        return 1
    parent_dir = os.path.dirname(target)
    os.makedirs(parent_dir, exist_ok=True)
    if not os.path.isdir(parent_dir):
        log.error("%s is not a directory.", parent_dir)
        return 2
    with tempfile.TemporaryDirectory(prefix="bigartm-", dir=tempdir) as tmpdir:
        log.info("Building bigartm/bigartm in %s...", tmpdir)
        execute("git clone --single-branch --depth=1 https://github.com/bigartm/bigartm .",
                tmpdir, log)
        cwd = os.path.join(tmpdir, "build")
        os.mkdir(cwd)
        execute("cmake -DCMAKE_BUILD_TYPE=Release -DBUILD_TESTING=OFF -DPYTHON=python3 ..",
                cwd, log)
        execute("make -j%d" % multiprocessing.cpu_count(), cwd, log)
        whl_path = glob.glob(os.path.join(tmpdir, "build/python/*.whl"))[0]
        execute("pip3 install \"%s\"" % whl_path, cwd, log)
        shutil.copyfile(os.path.join(cwd, "bin", "bigartm"), target)
        os.chmod(target, 0o777)
    log.info("Installed %s", os.path.abspath(target))



from modelforge.model import Model, split_strings, assemble_sparse_matrix, \
    merge_strings, disassemble_sparse_matrix
from modelforge.models import register_model


@register_model
class Cooccurrences(Model):
    """
    Co-occurrence matrix.
    """
    NAME = "co-occurrences"

    def construct(self, tokens, matrix):
        self._tokens = tokens
        self._matrix = matrix
        return self

    def _load_tree(self, tree):
        self.construct(tokens=split_strings(tree["tokens"]),
                       matrix=assemble_sparse_matrix(tree["matrix"]))

    def dump(self):
        return """Number of words: %d
First 10 words: %s
Matrix: shape: %s non-zero: %d""" % (
            len(self.tokens), self.tokens[:10], self.matrix.shape, self.matrix.getnnz())

    @property
    def tokens(self):
        """
        Returns the tokens in the order which corresponds to the matrix's rows and cols.
        """
        return self._tokens

    @property
    def matrix(self):
        """
        Returns the sparse co-occurrence matrix.
        """
        return self._matrix

    def __len__(self):
        """
        Returns the number of tokens in the model.
        """
        return len(self._tokens)

    def _generate_tree(self):
        return {"tokens": merge_strings(self.tokens),
                "matrix": disassemble_sparse_matrix(self.matrix)}



import sys

import numpy
from scipy.sparse import csr_matrix

from sourced.ml.models import Topics


def bigartm2asdf_entry(args):
    """
    BigARTM "readable" model -> Topics -> Modelforge ASDF.
    """
    tokens = []
    data = []
    indices = []
    indptr = [0]
    if args.input != "-":
        fin = open(args.input)
    else:
        fin = sys.stdin
    try:
        # the first line is the header
        fin.readline()
        for line in fin:
            items = line.split(";")
            tokens.append(items[0])
            nnz = 0
            for i, v in enumerate(items[2:]):
                if v == "0":
                    continue
                nnz += 1
                data.append(float(v))
                indices.append(i)
            indptr.append(indptr[-1] + nnz)
    finally:
        if args.input != "-":
            fin.close()
    data = numpy.array(data, dtype=numpy.float32)
    indices = numpy.array(indices, dtype=numpy.int32)
    matrix = csr_matrix((data, indices, indptr), shape=(len(tokens), len(items) - 2)).T
    Topics().construct(tokens, None, matrix).save(args.output)



from typing import Dict

from modelforge import Model, split_strings, merge_strings, register_model
import numpy


@register_model
class QuantizationLevels(Model):
    """
    This model contains quantization levels for multiple schemes (feature types).
    Every feature "class" (type, possible distinct value) corresponds to the numpy array
    with integer level borders. The size of each numpy array is (the number of levels + 1).
    """
    NAME = "quant"

    def construct(self, levels: Dict[str, Dict[str, numpy.ndarray]]):
        self._levels = levels
        return self

    @property
    def levels(self) -> Dict[str, Dict[str, numpy.ndarray]]:
        return self._levels

    def __len__(self):
        return len(self.levels)

    def _load_tree(self, tree):
        self._levels = {}
        for key, vals in tree["schemes"].items():
            classes = split_strings(vals["classes"])
            levels = vals["levels"]
            self.levels[key] = dict(zip(classes, numpy.split(levels, len(classes))))

    def _generate_tree(self):
        tree = {"schemes": {}}
        for key, vals in self.levels.items():
            tree["schemes"][key] = scheme = {}
            npartitions = len(next(iter(vals.values())))
            classes = [None for _ in range(len(vals))]
            scheme["levels"] = levels = numpy.zeros(len(vals) * npartitions, dtype=numpy.int32)
            for i, pair in enumerate(vals.items()):
                classes[i], levels[i * npartitions:(i + 1) * npartitions] = pair
            scheme["classes"] = merge_strings(classes)
        return tree

    def dump(self):
        return """Schemes: %s""" % (
            sorted((v[0], "%d@%d" % (len(v[1]), len(next(iter(v[1].values()))) - 1))
                   for v in self.levels.items()))



import codecs
import os

from sourced.ml.algorithms.uast_ids_to_bag import UastIds2Bag
from sourced.ml.extractors import BagsExtractor, register_extractor


class HashedTokenParser:
    def process_token(self, token):
        yield codecs.encode((hash(token) & 0xffffffffffffffff).to_bytes(8, "little"),
                            "hex_codec").decode()


class Literals2Bag(UastIds2Bag):
    """
    Converts a UAST to a bag-of-literals.
    """

    XPATH = "//*[@roleLiteral]"

    def __init__(self, token2index=None, token_parser=None):
        """
        :param token2index: The mapping from tokens to bag keys. If None, no mapping is performed.
        :param token_parser: Specify token parser if you want to use a custom one. \
            :class:'TokenParser' is used if it is not specified.
        """
        token_parser = HashedTokenParser() if token_parser is None else token_parser
        super().__init__(token2index, token_parser)


@register_extractor
class LiteralsBagExtractor(BagsExtractor):
    NAME = "lit"
    NAMESPACE = "l."
    OPTS = BagsExtractor.OPTS.copy()

    def __init__(self, docfreq_threshold=None, **kwargs):
        super().__init__(docfreq_threshold, **kwargs)
        self.id2bag = Literals2Bag(None, HashedTokenParser())

    def uast_to_bag(self, uast):
        if os.getenv("PYTHONHASHSEED", "random") == "random":
            raise RuntimeError("PYTHONHASHSEED must be set")
        return self.id2bag(uast)



import argparse
import logging

try:
    from sourced.ml.algorithms import swivel as swivel

    def mirror_tf_args(parser: argparse.ArgumentParser):
        """
        Copies the command line flags registered in swivel.py to an :class:
        `argparse.ArgumentParser`.
        :param parser: object to copy flags to.
        :return: None
        """
        types = {"string": str, "int": int, "float": float, "bool": bool}
        for flag in swivel.FLAGS.__dict__["__wrapped"].__dict__["__flags"].values():
            parser.add_argument(
                "--" + flag.name, default=flag.default, type=types[flag.flag_type()],
                help=flag.help)
except ImportError as e:
    logging.warning("Tensorflow is not installed, dependent functionality is unavailable.")

    def mirror_tf_args(parser: argparse.ArgumentParser):
        """Dummy handler to avoid errors in main."""
        pass


def run_swivel(args: argparse.Namespace):
    """
    Trains the Swivel model. Wraps swivel.py, taken from
    https://github.com/src-d/tensorflow-swivel/blob/master/swivel.py

    :param args: :class:`argparse.Namespace` identical to \
                 :class:`tf.app.flags`.
    :return: None
    """
    swivel.FLAGS = args
    logging.getLogger("tensorflow").handlers.clear()
    swivel.main(args)



import argparse
import json

from sourced.ml import extractors
from sourced.ml.transformers import Moder


def add_vocabulary_size_arg(my_parser: argparse.ArgumentParser):
    my_parser.add_argument(
        "-v", "--vocabulary-size", default=10000000, type=int,
        help="The maximum vocabulary size.")


def add_repo2_args(my_parser: argparse.ArgumentParser, quant=True):
    my_parser.add_argument(
        "-r", "--repositories", required=True,
        help="The path to the repositories.")
    my_parser.add_argument(
        "--graph", help="Write the tree in Graphviz format to this file.")
    my_parser.add_argument(
        "--pause", action="store_true",
        help="Do not terminate in the end - useful for inspecting Spark Web UI.")
    my_parser.add_argument(
        "--min-docfreq", default=1, type=int,
        help="The minimum document frequency of each feature.")
    add_vocabulary_size_arg(my_parser)
    my_parser.add_argument(
        "-l", "--languages", required=True, nargs="+", choices=(
            "Java", "Python", "JavaScript", "Ruby", "Bash"),
        help="The programming languages to analyse.")
    my_parser.add_argument(
        "--docfreq", required=True,
        help="[OUT] The path to the OrderedDocumentFrequencies model.")
    if quant:
        my_parser.add_argument(
            "--quant", help="[OUT] The path to the QuantizationLevels model.")


def add_feature_args(my_parser: argparse.ArgumentParser, required=True):
    my_parser.add_argument("-x", "--mode", choices=Moder.Options.__all__,
                           default="file", help="What to select for analysis.")
    my_parser.add_argument(
        "-f", "--feature", nargs="+",
        choices=[ex.NAME for ex in extractors.__extractors__.values()],
        required=required, help="The feature extraction scheme to apply.")
    for ex in extractors.__extractors__.values():
        for opt, val in ex.OPTS.items():
            my_parser.add_argument(
                "--%s-%s" % (ex.NAME, opt), default=val, type=json.loads,
                help="%s's kwarg" % ex.__name__)



import logging
import os
import sys

import numpy

from sourced.ml.algorithms import TokenParser
from sourced.ml.models import Id2Vec


def postprocess_id2vec(args):
    """
    Merges row and column embeddings produced by Swivel and writes the Id2Vec
    model.

    :param args: :class:`argparse.Namespace` with "swivel_data" \
                 and "result". The text files are read from \
                 `swivel_data` and the model is written to \
                 `result`.
    :return: None
    """
    log = logging.getLogger("postproc")
    log.info("Parsing the embeddings at %s...", args.swivel_data)
    tokens = []
    embeddings = []
    swd = args.swivel_data
    with open(os.path.join(swd, "row_embedding.tsv")) as frow:
        with open(os.path.join(swd, "col_embedding.tsv")) as fcol:
            for i, (lrow, lcol) in enumerate(zip(frow, fcol)):
                if i % 10000 == (10000 - 1):
                    sys.stdout.write("%d\r" % (i + 1))
                    sys.stdout.flush()
                prow, pcol = (l.split("\t", 1) for l in (lrow, lcol))
                assert prow[0] == pcol[0]
                tokens.append(prow[0][:TokenParser.MAX_TOKEN_LENGTH])
                erow, ecol = \
                    (numpy.fromstring(p[1], dtype=numpy.float32, sep="\t")
                     for p in (prow, pcol))
                embeddings.append((erow + ecol) / 2)
    log.info("Generating numpy arrays...")
    embeddings = numpy.array(embeddings, dtype=numpy.float32)
    log.info("Writing %s...", args.output)
    Id2Vec().construct(embeddings=embeddings, tokens=tokens).save(args.output)



from collections import defaultdict
from typing import Dict


class Uast2BagBase:
    """
    Base class to convert UAST to a bag of anything.
    """
    def __call__(self, uast) -> Dict[str, int]:
        """
        Inheritors must implement this function.

        :param uast: The UAST root node.
        """
        raise NotImplemented


class Uast2BagThroughSingleScan(Uast2BagBase):
    """
    Constructs the bag by doing a single tree traversal and turning every node into a string.
    """
    def __call__(self, uast):
        result = defaultdict(int)
        stack = [uast]
        while stack:
            node = stack.pop()
            stack.extend(node.children)
            result[self.node2key(node)] += 1
        return result

    def node2key(self, node) -> str:
        raise NotImplemented



import argparse
import logging
import os
import sys

from modelforge.logs import setup_logging

from sourced.ml.cmd_entries import bigartm2asdf_entry, dump_model, projector_entry, bow2vw_entry, \
    run_swivel, postprocess_id2vec, preprocess_id2vec, repos2coocc_entry, repos2df_entry, \
    repos2bow_entry
from sourced.ml.cmd_entries.args import add_repo2_args, add_feature_args, add_vocabulary_size_arg
from sourced.ml.cmd_entries.repos2bow import add_bow_args
from sourced.ml.cmd_entries.run_swivel import mirror_tf_args
from sourced.ml.utils import install_bigartm, add_engine_args


class ArgumentDefaultsHelpFormatterNoNone(argparse.ArgumentDefaultsHelpFormatter):
    """
    Pretty formatter of help message for arguments.
    It adds default value to the end if it is not None.
    """
    def _get_help_string(self, action):
        if action.default is None:
            return action.help
        return super()._get_help_string(action)


def get_parser() -> argparse.ArgumentParser:
    """
    Creates the cmdline argument parser.
    """

    parser = argparse.ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatterNoNone)
    parser.add_argument("--log-level", default="INFO",
                        choices=logging._nameToLevel,
                        help="Logging verbosity.")
    # Create and construct subparsers
    subparsers = parser.add_subparsers(help="Commands", dest="command")

    def add_parser(name, help):
        return subparsers.add_parser(
            name, help=help, formatter_class=ArgumentDefaultsHelpFormatterNoNone)

    # ------------------------------------------------------------------------
    repos2bow_parser = add_parser(
        "repos2bow", "Convert source code to the bag-of-words model.")
    repos2bow_parser.set_defaults(handler=repos2bow_entry)
    add_repo2_args(repos2bow_parser)
    add_engine_args(repos2bow_parser)
    add_bow_args(repos2bow_parser)
    add_feature_args(repos2bow_parser)
    # ------------------------------------------------------------------------
    repos2df_parser = add_parser(
        "repos2df", "Calculate document frequencies of features extracted from source code.")
    repos2df_parser.set_defaults(handler=repos2df_entry)
    add_repo2_args(repos2df_parser)
    add_engine_args(repos2df_parser)
    add_feature_args(repos2df_parser)
    # ------------------------------------------------------------------------
    repos2coocc_parser = add_parser(
        "repos2coocc", "Convert source code to the sparse co-occurrence matrix of identifiers.")
    add_engine_args(repos2coocc_parser)
    add_repo2_args(repos2coocc_parser, quant=False)
    repos2coocc_parser.add_argument(
        "-o", "--output", required=True,
        help="[OUT] Path to the Cooccurrences model.")
    repos2coocc_parser.add_argument(
        "--split-stem", default=False, action="store_true",
        help="Split Tokens to parts (ThisIs_token -> ['this', 'is', 'token']).")
    repos2coocc_parser.set_defaults(handler=repos2coocc_entry)
    # ------------------------------------------------------------------------
    preproc_parser = add_parser(
        "id2vec_preproc", "Convert a sparse co-occurrence matrix to the Swivel shards.")
    preproc_parser.set_defaults(handler=preprocess_id2vec)
    add_vocabulary_size_arg(preproc_parser)
    preproc_parser.add_argument("-s", "--shard-size", default=4096, type=int,
                                help="The shard (submatrix) size.")
    preproc_parser.add_argument(
        "--docfreq", default=None,
        help="[IN] Path to the pre-calculated document frequencies in asdf format "
             "(DF in TF-IDF).")
    preproc_parser.add_argument(
        "-i", "--input",
        help="Concurrence model produced by repos2coocc.")
    preproc_parser.add_argument("-o", "--output", required=True, help="Output directory.")
    # ------------------------------------------------------------------------
    train_parser = add_parser(
        "id2vec_train", "Train identifier embeddings using Swivel.")
    mirror_tf_args(train_parser)
    train_parser.set_defaults(handler=run_swivel)
    # ------------------------------------------------------------------------
    id2vec_postproc_parser = add_parser(
        "id2vec_postproc",
        "Combine row and column embeddings produced by Swivel and write them to an .asdf.")
    id2vec_postproc_parser.set_defaults(handler=postprocess_id2vec)
    id2vec_postproc_parser.add_argument(
        "-i", "--swivel-data", required=True,
        help="Folder with swivel batches input data. You can get it using repos2coocc subcommand.")
    id2vec_postproc_parser.add_argument(
        "-o", "--output", required=True,
        help="Output directory for embedding data.")
    # ------------------------------------------------------------------------
    id2vec_project_parser = add_parser(
        "id2vec_project", "Present id2vec model in Tensorflow Projector.")
    id2vec_project_parser.set_defaults(handler=projector_entry)
    id2vec_project_parser.add_argument("-i", "--input", required=True,
                                       help="id2vec model to present.")
    id2vec_project_parser.add_argument("-o", "--output", required=True,
                                       help="Projector output directory.")
    id2vec_project_parser.add_argument("--docfreq", help="docfreq model to pick the most "
                                                         "significant identifiers.")
    id2vec_project_parser.add_argument("--no-browser", action="store_true",
                                       help="Do not open the browser.")
    # ------------------------------------------------------------------------
    bow2vw_parser = add_parser(
        "bow2vw", "Convert a bag-of-words model to the dataset in Vowpal Wabbit format.")
    bow2vw_parser.set_defaults(handler=bow2vw_entry)
    bow2vw_parser.add_argument(
        "--bow", help="URL or path to a bag-of-words model.")
    bow2vw_parser.add_argument(
        "--id2vec", help="URL or path to the identifier embeddings.")
    bow2vw_parser.add_argument(
        "-o", "--output", required=True, help="Path to the output file.")
    # ------------------------------------------------------------------------
    bigartm_postproc_parser = add_parser(
        "bigartm2asdf", "Convert a human-readable BigARTM model to Modelforge format.")
    bigartm_postproc_parser.set_defaults(handler=bigartm2asdf_entry)
    bigartm_postproc_parser.add_argument("input")
    bigartm_postproc_parser.add_argument("output")
    # ------------------------------------------------------------------------
    bigartm_parser = add_parser(
        "bigartm", "Install bigartm/bigartm to the current working directory.")
    bigartm_parser.set_defaults(handler=install_bigartm)
    bigartm_parser.add_argument(
        "--tmpdir", help="Store intermediate files in this directory instead of /tmp.")
    bigartm_parser.add_argument("--output", default=os.getcwd(), help="Output directory.")
    # ------------------------------------------------------------------------
    dump_parser = add_parser("dump", "Dump a model to stdout.")
    dump_parser.set_defaults(handler=dump_model)
    dump_parser.add_argument(
        "input", help="Path to the model file, URL or UUID.")
    dump_parser.add_argument("--gcs", default=None, dest="gcs_bucket",
                             help="GCS bucket to use.")

    return parser


def main():
    """
    Creates all the argparse-rs and invokes the function from set_defaults().

    :return: The result of the function from set_defaults().
    """

    parser = get_parser()
    args = parser.parse_args()
    args.log_level = logging._nameToLevel[args.log_level]
    setup_logging(args.log_level)
    try:
        handler = args.handler
    except AttributeError:
        def print_usage(_):
            parser.print_usage()

        handler = print_usage
    return handler(args)


if __name__ == "__main__":
    sys.exit(main())



from sourced.ml.models.bow import BOW
from sourced.ml.models.coocc import Cooccurrences
from sourced.ml.models.df import DocumentFrequencies
from sourced.ml.models.ordered_df import OrderedDocumentFrequencies
from sourced.ml.models.id2vec import Id2Vec
from sourced.ml.models.topics import Topics
from sourced.ml.models.quant import QuantizationLevels



import logging
from uuid import uuid4

from sourced.ml.extractors import IdentifiersBagExtractor
from sourced.ml.models import OrderedDocumentFrequencies
from sourced.ml.transformers import Ignition, HeadFiles, UastExtractor, Cacher, UastDeserializer, \
    CooccConstructor, CooccModelSaver, BagFeatures2DocFreq, Uast2BagFeatures, Counter, \
    UastRow2Document
from sourced.ml.utils import create_engine
from sourced.ml.utils.engine import pipeline_graph, pause


@pause
def repos2coocc_entry(args):
    log = logging.getLogger("repos2coocc")
    engine = create_engine("repos2coocc-%s" % uuid4(), **args.__dict__)
    id_extractor = IdentifiersBagExtractor(docfreq_threshold=args.min_docfreq,
                                           split_stem=args.split_stem)

    ignition = Ignition(engine, explain=args.explain)
    uast_extractor = ignition \
        .link(HeadFiles()) \
        .link(UastExtractor(languages=args.languages)) \
        .link(UastRow2Document()) \
        .link(Cacher.maybe(args.persist))
    log.info("Extracting UASTs...")
    ndocs = uast_extractor.link(Counter()).execute()
    log.info("Number of documents: %d", ndocs)
    uast_extractor = uast_extractor.link(UastDeserializer())

    df = uast_extractor \
        .link(Uast2BagFeatures([id_extractor])) \
        .link(BagFeatures2DocFreq()) \
        .execute()

    log.info("Writing document frequency model to %s...", args.docfreq)
    df_model = OrderedDocumentFrequencies() \
        .construct(ndocs, df) \
        .prune(args.min_docfreq) \
        .greatest(args.vocabulary_size) \
        .save(args.docfreq)

    token2index = engine.session.sparkContext.broadcast(df_model.order)
    uast_extractor \
        .link(CooccConstructor(token2index=token2index,
                               token_parser=id_extractor.id2bag.token_parser,
                               namespace=id_extractor.NAMESPACE)) \
        .link(CooccModelSaver(args.output, df_model)) \
        .execute()
    pipeline_graph(args, log, ignition)



from typing import Union

from pyspark import StorageLevel, Row, RDD
from pyspark.sql import DataFrame

from sourced.ml.transformers.transformer import Transformer
from sourced.ml.utils import EngineConstants


class Sampler(Transformer):
    """
    Wraps `sample()` function from pyspark Dataframe.
    """
    def __init__(self, with_replacement=False, fraction=0.05, seed=42, **kwargs):
        super().__init__(**kwargs)
        self.with_replacement = with_replacement
        self.fraction = fraction
        self.seed = seed

    def __call__(self, head):
        return head.sample(self.with_replacement, self.fraction, self.seed)


class Collector(Transformer):
    def __call__(self, head: RDD):
        return head.collect()


class First(Transformer):
    def __call__(self, head: RDD):
        return head.first()


class Identity(Transformer):
    def __call__(self, head: RDD):
        return head


class Cacher(Transformer):
    def __init__(self, persistence, **kwargs):
        super().__init__(**kwargs)
        self.persistence = getattr(StorageLevel, persistence)
        self.head = None
        self.trace = None

    def __getstate__(self):
        state = super().__getstate__()
        state["head"] = None
        state["trace"] = None
        return state

    def __call__(self, head: RDD):
        if self.head is None or self.trace != self.path():
            self.head = head.persist(self.persistence)
            self.trace = self.path()
        return self.head

    @staticmethod
    def maybe(persistence):
        if persistence is not None:
            return Cacher(persistence)
        else:
            return Identity()


class Ignition(Transformer):
    def __init__(self, engine, **kwargs):
        super().__init__(**kwargs)
        self.engine = engine

    def __getstate__(self):
        state = super().__getstate__()
        del state["engine"]
        return state

    def __call__(self, _):
        return self.engine


class HeadFiles(Transformer):
    def __call__(self, engine):
        return engine.repositories.references.head_ref.commits.tree_entries.blobs


class Counter(Transformer):
    def __init__(self, distinct=False, approximate=False, **kwargs):
        super().__init__(**kwargs)
        self.distinct = distinct
        self.approximate = approximate

    def __call__(self, head: RDD):
        if self.distinct and not self.approximate:
            head = head.distinct()
        if self.explained:
            self._log.info("toDebugString():\n%s", head.toDebugString().decode())
        if not self.approximate or not self.distinct:
            return head.count()
        return head.countApproxDistinct()


class UastExtractor(Transformer):
    def __init__(self, languages: Union[list, tuple], **kwargs):
        super().__init__(**kwargs)
        self.languages = languages

    def __call__(self, files: DataFrame) -> DataFrame:
        files = files.dropDuplicates(("blob_id",)).filter("is_binary = 'false'")
        classified = files.classify_languages()
        lang_filter = classified.lang == self.languages[0]
        for lang in self.languages[1:]:
            lang_filter |= classified.lang == lang
        filtered_by_lang = classified.filter(lang_filter)
        from pyspark.sql import functions
        uasts = filtered_by_lang.extract_uasts().where(functions.size(functions.col("uast")) > 0)
        return uasts


class FieldsSelector(Transformer):
    def __init__(self, fields: Union[list, tuple], **kwargs):
        super().__init__(**kwargs)
        self.fields = fields

    def __call__(self, df: DataFrame) -> DataFrame:
        res = df.select(self.fields)
        if self.explained:
            self._log.info("toDebugString():\n%s", res.rdd.toDebugString().decode())
        return res


class ParquetSaver(Transformer):
    def __init__(self, save_loc, **kwargs):
        super().__init__(**kwargs)
        self.save_loc = save_loc

    def __call__(self, df: DataFrame):
        if self.explained:
            self._log.info("toDebugString():\n%s", df.rdd.toDebugString().decode())
        df.write.parquet(self.save_loc)


class ParquetLoader(Transformer):
    def __init__(self, session, paths, **kwargs):
        super().__init__(**kwargs)
        self.session = session
        self.paths = paths

    def __getstate__(self):
        state = super().__getstate__()
        del state["session"]
        return state

    def __call__(self, _):
        if isinstance(self.paths, (list, tuple)):
            return self.session.read.parquet(*self.paths)
        if isinstance(self.paths, str):
            return self.session.read.parquet(self.paths)
        raise ValueError


class UastDeserializer(Transformer):
    def __setstate__(self, state):
        super().__setstate__(state)
        from bblfsh import Node
        self.parse_uast = Node.FromString

    def __call__(self, rows: RDD) -> RDD:
        return rows.flatMap(self.deserialize_uast)

    def deserialize_uast(self, row: Row):
        if not row[EngineConstants.Columns.Uast]:
            return
        row_dict = row.asDict()
        row_dict[EngineConstants.Columns.Uast] = [
            self.parse_uast(uast) for uast in row[EngineConstants.Columns.Uast]]
        yield Row(**row_dict)
